@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2000-2013 Erwin Waterlander
+Copyright (C) 2000-2016 Erwin Waterlander
 
 Ideas and source code of NCD (Ninux Czo Directory) have been
 used in the WCD graphical interface.
@@ -116,7 +116,6 @@ struct wcdwin /* structure with window information */
 {
    WINDOW *scrollWin;
    int scrollWinHeight;
-   int scrollWinLen;
    WINDOW *inputWin;
    dirnode curNode;
    char str[WCD_MAX_INPSTR];
@@ -139,7 +138,6 @@ void dataRefresh(int ydiff, int init);
 #  define CAN_RESIZE 0
 #endif
 
-#if CAN_RESIZE
 /* The most portable way of resizing seems to be
  * to just rebuild the windows from scratch */
 void ioResize()
@@ -149,15 +147,14 @@ void ioResize()
    refresh();  /* start curses */
 
    wcd_cwin.scrollWinHeight = LINES - INPUT_WIN_HEIGHT;
-   wcd_cwin.scrollWinLen = COLS;
 
    /* free resources */
    delwin(wcd_cwin.scrollWin);
    delwin(wcd_cwin.inputWin);
 
    /* create new windows */
-   wcd_cwin.scrollWin = newpad(wcd_cwin.scrollWinHeight,COLS);
-   wcd_cwin.inputWin  = newpad(INPUT_WIN_HEIGHT,COLS);
+   wcd_cwin.scrollWin = newwin(wcd_cwin.scrollWinHeight,COLS,0,0);
+   wcd_cwin.inputWin = newwin(INPUT_WIN_HEIGHT,COLS,wcd_cwin.scrollWinHeight,0);
 
    if (wcd_cwin.graphics_mode & WCD_GRAPH_COLOR)
    {
@@ -170,6 +167,7 @@ void ioResize()
    dataRefresh(0, 1);
 }
 
+#if CAN_RESIZE
 void signalSigwinch (int sig)
 {
   ioResize ();
@@ -234,7 +232,7 @@ char* getTreeLine(dirnode d, int y, int *y_orig, dirnode curNode, c3po_bool fold
    static text line = NULL;
    static text tline = NULL;
    dirnode n;
-   size_t i,len,clen;
+   size_t len,clen;
 
    if (d == NULL)
       return(NULL);
@@ -305,6 +303,7 @@ char* getTreeLine(dirnode d, int y, int *y_orig, dirnode curNode, c3po_bool fold
          }
          else
          {
+            size_t i;
             for (i=0;i<clen;i++)
                tline[i] = ' ';
             tline[clen] = '\0';
@@ -336,6 +335,7 @@ char* getTreeLine(dirnode d, int y, int *y_orig, dirnode curNode, c3po_bool fold
             }
             else
             {
+               size_t i;
                for (i=0;i<clen;i++)
                   tline[i] = ' ';
                tline[clen] = '\0';
@@ -344,6 +344,7 @@ char* getTreeLine(dirnode d, int y, int *y_orig, dirnode curNode, c3po_bool fold
          }
          else
          {
+            size_t i;
             for (i=0;i<clen;i++)
                tline[i] = ' ';
             tline[clen] = '\0';
@@ -418,6 +419,8 @@ void dumpTreeLine(dirnode d, const int *graphics_mode)
 
       y = dirnodeGetY(d);
       l = getTreeLine(d,y,&y,NULL,false, graphics_mode);
+      if (l == NULL)
+         return;
       while ( *l != '\0')
          {
             switch(*l)
@@ -496,14 +499,13 @@ void dumpTree(dirnode d, const int *graphics_mode)
 
 void setXYTree(dirnode d, const int *graphics_mode)
 {
-   size_t index, len, size;
-   int x;
-   static int y;
-   dirnode n;
-
 
    if(dirHasSubdirs(d) eq true)
    {
+      int x;
+      static int y;
+      size_t index, len, size;
+      dirnode n;
       if (*graphics_mode & WCD_GRAPH_COMPACT) /* compact tree */
       {
          len = str_columns(dirnodeGetName(d));
@@ -560,29 +562,29 @@ c3po_bool eqTextDirnode(text t, dirnode d)
 
 size_t inDirnode(text t, dirnode d)
 {
-   size_t index, size;
+   size_t index;
 
    index = 0;
 
-   if(dirHasSubdirs(d) eq true)
+   while(index < d->size)
    {
-      size = getSizeOfDirnode(d);
-      while(index < size)
-      {
-         if (eqTextDirnode(t,elementAtDirnode(index,d)) eq true)
-            return index;
-         index++;
-      }
+      if (strcmp(t,(d->subdirs[index])->name) == 0)
+         return index;
+      ++index;
    }
    return (size_t) -1;
 }
 
+/*
+ * A lot of time is spend in addPath(). addPath() has been made faster by
+ * using pointers directly instead of the dirnode API functions in addPath()
+ * and inDirnode(). This removes overhead due to pointer checks.
+ */
 void addPath(text path, dirnode d)
 {
    char *s;
    dirnode n,n_up;
    text t;
-   size_t index;
 #if (defined(_WIN32) || defined(__CYGWIN__))
    static char buf[DD_MAXPATH] = "//" ;
 
@@ -604,7 +606,7 @@ void addPath(text path, dirnode d)
 
    if (s != NULL)
    {
-      index = inDirnode(s,d);
+      size_t index = inDirnode(s,d);
       if (index == (size_t) -1)
       {
          n = dirnodeNew(d,NULL,NULL);
@@ -621,12 +623,132 @@ void addPath(text path, dirnode d)
       }
       else
       {
-         n = elementAtDirnode(index,d);
+         n = d->subdirs[index];
       }
       addPath(NULL,n);
    }
 }
 
+/************************************************************************
+ *
+ *  swap_dirnodes(), ssort_dirnode() and sort_dirnode()
+ *
+ *  Sort the sub dirnodes of a dirnode alphabeticly
+ *
+ ************************************************************************/
+
+void swap_dirnodes(dirnode* subdirs, int i, int j)
+{  dirnode temp, parent;
+
+   parent = (subdirs[i])->parent;
+
+   temp = subdirs[i];
+   subdirs[i] = subdirs[j];
+   subdirs[j] = temp;
+
+   if (i==0)
+     (subdirs[i])->up = NULL;
+   else {
+     (subdirs[i])->up = subdirs[i-1];
+     (subdirs[i-1])->down = subdirs[i];
+   }
+   if (i== (int)parent->size -1)
+     (subdirs[i])->down = NULL;
+   else {
+     (subdirs[i])->down = subdirs[i+1];
+     (subdirs[i+1])->up = subdirs[i];
+   }
+
+   if (j==0)
+     (subdirs[j])->up = NULL;
+   else {
+     (subdirs[j])->up = subdirs[j-1];
+     (subdirs[j-1])->down = subdirs[j];
+   }
+   if (j== (int)parent->size -1)
+     (subdirs[j])->down = NULL;
+   else {
+     (subdirs[j])->down = subdirs[j+1];
+     (subdirs[j+1])->up = subdirs[j];
+   }
+}
+
+void ssort_dirnode (dirnode* subdirs, int left, int right)
+{
+ int i, last;
+#if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
+ static wchar_t wstr_left[DD_MAXPATH];
+ static wchar_t wstr_right[DD_MAXPATH];
+#endif
+ text t_left, t_right;
+
+  if (left >= right) return; /* fewer than 2 elements */
+
+  swap_dirnodes(subdirs, left, (left + right)/2);
+  last = left;
+
+  for (i = left+1; i <=right; i++)
+  {
+   t_left  = (subdirs[left])->name;
+   t_right = (subdirs[i])->name;
+#if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
+   size_t len1 = MBSTOWCS(wstr_left, t_left, (size_t)DD_MAXPATH);
+   size_t len2 = MBSTOWCS(wstr_right, t_right, (size_t)DD_MAXPATH);
+   if ((len1 == (size_t)(-1)) || (len2 == (size_t)(-1)))
+   {
+      /* Erroneous multi-byte sequence */
+      /* Try 8 bit characters */
+#  ifdef ENABLE_NLS
+      if  (strcoll(t_right,t_left)<0)
+#  else
+      if  (strcmp(t_right,t_left)<0)
+#  endif
+         swap_dirnodes(subdirs, ++last, i);
+   } else {
+      if  (wcscoll(wstr_right,wstr_left)<0)
+         swap_dirnodes(subdirs, ++last, i);
+   }
+#else
+#  ifdef ENABLE_NLS
+  if  (strcoll(t_right,t_left)<0)
+#  else
+  if  (strcmp(t_right,t_left)<0)
+#  endif
+     swap_dirnodes(subdirs, ++last, i);
+#endif
+  }
+
+  swap_dirnodes(subdirs, left, last);
+  ssort_dirnode(subdirs, left, last-1);
+  ssort_dirnode(subdirs, last+1, right);
+}
+
+void sort_subdirs(dirnode d)
+{
+  ssort_dirnode(d->subdirs,0,(int)(d->size)-1);
+}
+
+/*
+ * Sort the tree alphabeticly
+ */
+
+void sortTree(dirnode d)
+{
+   size_t index;
+   dirnode n;
+
+   sort_subdirs(d);
+   index = 0;
+   while(index < d->size)
+   {
+      n = d->subdirs[index];
+      sortTree(n);
+      ++index;
+   }
+}
+
+/*****************************************************/
+
 void buildTreeFromNameset(nameset set, dirnode d)
 {
    size_t i;
@@ -664,23 +786,59 @@ void buildTreeFromNameset(nameset set, dirnode d)
  *
  ***********************************************************/
 
-void buildTreeFromFile(char *filename, dirnode d)
+
+/*****************************************************/
+
+/*
+ * When reading complete tree files, it is faster to build the
+ * dirnode tree directly from file, instead of reading the tree files
+ * in a nameset first. Building the nameset costs a lot of reallocs.
+ *
+ * Each time this function is called the file is appended to the dirnode.
+ *
+ * Also sorting a dirnode tree is faster, because we only need to compare
+ * directory names instead of whole paths.
+ *
+ * Most time is spend in addPath(). addPath() has been made faster by
+ * using pointers directly instead of the dirnode API functions.
+ *
+ */
+
+void buildTreeFromFile(char* filename, dirnode d, int quiet)
 {
-   nameset dirs;
+   FILE *infile;
+   int bomtype, line_nr=1;
 
-   dirs = namesetNew();
-
-   if(dirs != NULL)
-      read_treefile(filename,dirs,0);
-   else
+   if ((filename == NULL)||(d == NULL))
       return;
 
-   if (d == NULL)
-      return;
+   /* open treedata-file */
+   if  ((infile = wcd_fopen_bom(filename,"rb", quiet, &bomtype)) != NULL)
+   {
+       while (!feof(infile) && !ferror(infile))  /* parse the file */
+       {
+          int len;
+          char line[DD_MAXPATH];
 
-   buildTreeFromNameset(dirs,d);
-
-   freeNameset(dirs, 1);
+          len = read_treefile_line(line,infile,filename,&line_nr, bomtype);
+          if (ferror(infile)) {
+             wcd_read_error(filename);
+             wcd_fclose(infile, filename, "r", "buildTreeFromFile: ");
+             return;
+          }
+          ++line_nr;
+          if (len > 0 )
+          {
+             wcd_fixpath(line,sizeof(line));
+             addPath(line,d);
+          }
+       }
+       wcd_fclose(infile, filename, "r", "buildTreeFromFile: ");
+   } else {
+      if (!quiet) {
+         wcd_read_error(filename);
+      }
+   }
 }
 
 /*****************************************************/
@@ -713,7 +871,6 @@ dirnode searchNodeForDir(text path, dirnode d, dirnode rNode)
 {
    char *s;
    dirnode n;
-   size_t index;
    dirnode rootnode;
 #if (defined(_WIN32) || defined(__CYGWIN__))
    static char buf[DD_MAXPATH] = "//" ;
@@ -747,7 +904,7 @@ dirnode searchNodeForDir(text path, dirnode d, dirnode rNode)
 
    if (s != NULL)
    {
-      index = inDirnode(s,d);
+      size_t index = inDirnode(s,d);
       if (index == (size_t) -1)
       {
          return(d);
@@ -1276,7 +1433,7 @@ int validSearchDir(char *dir, dirnode curNode, int exact, int ignore_case, int i
     * */
 
    strcpy(path,"*");
-   strcat(path,dir);
+   wcd_strncat(path,dir,sizeof(path));
    if (exact == 0)
       strcat(path,"*");
 
@@ -1308,6 +1465,8 @@ dirnode findDirInCiclePrev(char *dir, dirnode curNode, int exact, int ignore_cas
    dirnode node;
    int valid;
 
+   if (curNode == NULL)
+      return NULL;
    node = curNode;
 
    do{
@@ -1325,6 +1484,8 @@ dirnode findDirInCicle(char *dir, dirnode curNode, int exact, int ignore_case, i
    dirnode node;
    int valid;
 
+   if (curNode == NULL)
+      return NULL;
    node = curNode;
 
    do{
@@ -1387,21 +1548,17 @@ int wcd_wcswidth(const wchar_t *pwcs, size_t n)
 void updateLine(WINDOW *win, dirnode n, int i, int y, dirnode curNode, int xoffset)
 {
    wcd_uchar *s;
-   size_t len;
-   int j;
-#if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
-   static wchar_t wstr[DD_MAXPATH];
-   int width, c;
-#endif
 
    s = (wcd_uchar *)getTreeLine(getLastNodeInLevel(n,i),i,&i,curNode,false, &wcd_cwin.graphics_mode);
 
    if (s != NULL)
    {
+      int j;
 #if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
-      len = MBSTOWCS(wstr,(char *)s,(size_t)DD_MAXPATH); /* number of wide characters */
+      static wchar_t wstr[DD_MAXPATH];
+      size_t len = MBSTOWCS(wstr,(char *)s,(size_t)DD_MAXPATH); /* number of wide characters */
 #else
-      len = strlen((char *)s);
+      size_t len = strlen((char *)s);
 #endif
       wmove(win,y,0);
     /*  if (len == (size_t) -1 )
@@ -1473,6 +1630,7 @@ void updateLine(WINDOW *win, dirnode n, int i, int y, dirnode curNode, int xoffs
       }
       else
       {
+         int width, c;
          c = 0; /* count width from beginning of string. */
          j = 0;
          while ((j<(int)len)&&(c<xoffset))
@@ -1482,7 +1640,7 @@ void updateLine(WINDOW *win, dirnode n, int i, int y, dirnode curNode, int xoffs
          }
          /* xoffset is horizontal offset measured in nr. of columns.
           * This may be at the middle of a double width character. */
-         if (( c > xoffset ) && ( wcd_wcwidth(wstr[j-1]) == 2))
+         if (( c > xoffset ) && (j>0) && ( wcd_wcwidth(wstr[j-1]) == 2))
          {
             /* Last character skipped was a double width character.
              * Insert a space. */
@@ -1656,14 +1814,10 @@ char *getZoomStackPath(dirnode stack)
 /************************************************************/
 void dataRefresh(int ydiff, int init)
 {
-  int i, yoffset, xo, len;
+  int i, yoffset, len;
   static int xoffset = 0;  /* Horizontal offset in number of columns */
   static int yposition = -1;  /* -1 : not initialized */
   wcd_uchar *s;
-#if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
-  static wchar_t wstr[DD_MAXPATH];
-  int width;
-#endif
 
   /*
      yoffset is the y-offset in the whole tree of the current node.
@@ -1673,8 +1827,8 @@ void dataRefresh(int ydiff, int init)
   if ((yposition < 0)|| init)
      yposition = (wcd_cwin.scrollWinHeight /2);
 
-  werase(wcd_cwin.scrollWin);
-  werase(wcd_cwin.inputWin);
+  WCD_WCLEAR(wcd_cwin.scrollWin);
+  WCD_WCLEAR(wcd_cwin.inputWin);
 
   if (ydiff < 0 )
   {
@@ -1701,7 +1855,7 @@ void dataRefresh(int ydiff, int init)
   /* len is total nr of colums of current node plus 3 */
   if (len > COLS)
   {
-    xo = len - COLS;
+    int xo = len - COLS;
     if ((xo > xoffset) || (xoffset > (dirnodeGetX(wcd_cwin.curNode)-1)))
     {
        xoffset = xo;
@@ -1727,6 +1881,10 @@ void dataRefresh(int ydiff, int init)
 
   if (s != NULL)
   {
+#if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
+   static wchar_t wstr[DD_MAXPATH];  /* Declarations at beginning of scope for Watcom C */
+   int width;
+#endif
     wmove(wcd_cwin.inputWin, 1, 0);
 #if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
    len = (int)MBSTOWCS(wstr,(char *)s,(size_t)DD_MAXPATH); /* number of wide characters */
@@ -1781,15 +1939,15 @@ void dataRefresh(int ydiff, int init)
 #endif
   }
 
-  prefresh(wcd_cwin.scrollWin, 0, 0, 0, 0, wcd_cwin.scrollWinHeight-1, COLS-1);
-  prefresh(wcd_cwin.inputWin, 0, 0, wcd_cwin.scrollWinHeight, 0,
-                          wcd_cwin.scrollWinHeight+INPUT_WIN_HEIGHT-1, COLS-1);
+  wrefresh(wcd_cwin.scrollWin);
+  wrefresh(wcd_cwin.inputWin);
+
 }
 
 /************************************************************/
 void showHelp(WINDOW *win, int height)
 {
-   werase(win);
+   WCD_WCLEAR(win);
    if (height > 21)
    {
       wcd_mvwaddstr(win, 0,0,_("NAVIGATION MODE (1/2):"));
@@ -1809,9 +1967,9 @@ void showHelp(WINDOW *win, int height)
       wcd_mvwaddstr(win,14,0,_("d                 go half page down"));
       wcd_mvwaddstr(win,16,0,_("Press any key."));
 
-      prefresh(win,0,0,0,0,height-1,COLS-1);
+      wrefresh(win);
       getch();
-      werase(win);
+      WCD_WCLEAR(win);
 
       wcd_mvwaddstr(win, 0,0,_("NAVIGATION MODE (2/2):"));
       wcd_mvwaddstr(win, 1,0,_("A                 switch alternative tree navigation on/off"));
@@ -1822,12 +1980,13 @@ void showHelp(WINDOW *win, int height)
       wcd_mvwaddstr(win, 6,0,_("/                 search forward"));
       wcd_mvwaddstr(win, 7,0,_("?                 search backward"));
       wcd_mvwaddstr(win, 8,0,_("n                 repeat last / or ? search"));
-      wcd_mvwaddstr(win, 9,0,_("<Enter>           select directory"));
-      wcd_mvwaddstr(win,11,0,_("Press any key."));
+      wcd_mvwaddstr(win, 9,0,_("CTRL-l or F5      redraw screen"));
+      wcd_mvwaddstr(win,10,0,_("<Enter>           select directory"));
+      wcd_mvwaddstr(win,12,0,_("Press any key."));
 
-      prefresh(win,0,0,0,0,height-1,COLS-1);
+      wrefresh(win);
       getch();
-      werase(win);
+      WCD_WCLEAR(win);
 
       wcd_mvwaddstr(win, 0,0,_("SEARCH MODE with wildcard and subdir support:"));
       wcd_mvwaddstr(win, 1,0,_("<Left>            go left"));
@@ -1848,9 +2007,9 @@ void showHelp(WINDOW *win, int height)
       wcd_mvwaddstr(win,16,0,_("<Enter>           select directory"));
       wcd_mvwaddstr(win,18,0,_("Press any key."));
 
-      prefresh(win,0,0,0,0,height-1,COLS-1);
+      wrefresh(win);
       getch();
-      werase(win);
+      WCD_WCLEAR(win);
 
       wcd_mvwaddstr(win, 0,0,_("ZOOMING:"));
       wcd_mvwaddstr(win, 1,0,_("z or i or CTRL-i  zoom in"));
@@ -1869,7 +2028,7 @@ void showHelp(WINDOW *win, int height)
       wcd_mvwaddstr(win, 0,0,_("Screenheight must be > 21 for help."));
 
 
-   prefresh(win,0,0,0,0,height-1,COLS-1);
+   wrefresh(win);
    getch();
 }
 
@@ -1975,14 +2134,13 @@ void setFold(dirnode n, c3po_bool f, int *ymax)
  ****************************************************************/
 void setFold_tree(dirnode d, c3po_bool *f)
 {
-   size_t index,size;
    dirnode n;
 
    if(dirnodeHasSubdirs(d) eq true) /* only (un)fold directories that have subdirs */
    {
+      size_t index = 0;
+      size_t size = getSizeOfDirnode(d);
       d->fold = *f;
-      index = 0;
-      size = getSizeOfDirnode(d);
       while(index < size)
       {
          n = elementAtDirnode(index,d);
@@ -2096,8 +2254,14 @@ char *selectANode(dirnode tree, int *use_HOME, int ignore_case, int graphics_mod
 
    ptr=NULL;
 
-   getCurPath(curPath, (size_t)DD_MAXPATH, use_HOME);
+   if (getCurPath(curPath, (size_t)DD_MAXPATH, use_HOME) == NULL)
+      return NULL;
    wcd_cwin.curNode = locatePathOrSo(curPath,tree);
+   if (wcd_cwin.curNode == NULL)
+   {
+      print_error(_("Cannot find the current path %s in the directory tree.\n"), curPath);
+      return NULL;
+   }
 
    wcd_cwin.zoomStack = dirnodeNew(NULL,NULL,NULL);
 
@@ -2130,10 +2294,8 @@ char *selectANode(dirnode tree, int *use_HOME, int ignore_case, int graphics_mod
 
 
    wcd_cwin.scrollWinHeight = LINES - INPUT_WIN_HEIGHT;
-   wcd_cwin.scrollWinLen = COLS;
-   wcd_cwin.scrollWin = newpad(wcd_cwin.scrollWinHeight,COLS);
-
-   wcd_cwin.inputWin = newpad(INPUT_WIN_HEIGHT,COLS);
+   wcd_cwin.scrollWin = newwin(wcd_cwin.scrollWinHeight,COLS,0,0);
+   wcd_cwin.inputWin = newwin(INPUT_WIN_HEIGHT,COLS,wcd_cwin.scrollWinHeight,0);
 
    if (wcd_cwin.graphics_mode & WCD_GRAPH_COLOR)
    {
@@ -2487,6 +2649,10 @@ char *selectANode(dirnode tree, int *use_HOME, int ignore_case, int graphics_mod
       case KEY_F (1):
             showHelp(wcd_cwin.scrollWin,wcd_cwin.scrollWinHeight);
          break;
+      case KEY_F (5):
+      case Key_CTRL ('l'):
+            ioResize();
+         break;
       case Key_CTRL ('i'):
             pushZoom(wcd_cwin.zoomStack,wcd_cwin.curNode,&ymax);
          break;
@@ -2531,7 +2697,7 @@ char *selectANode(dirnode tree, int *use_HOME, int ignore_case, int graphics_mod
 #endif
      break;
      default:
-      if ((wcd_cwin.mode == WCD_SEARCH) && (n < WCD_MAX_INPSTR))
+      if ((wcd_cwin.mode == WCD_SEARCH) && (n < (WCD_MAX_INPSTR -1)))
       {
 #if defined(WCD_UNICODE) || defined(WCD_WINDOWS)
          wcd_cwin.wstr[n] = (wchar_t)ch;
@@ -2560,6 +2726,11 @@ char *selectANode(dirnode tree, int *use_HOME, int ignore_case, int graphics_mod
      break;
     }
 
+    if (wcd_cwin.curNode == NULL)
+    {
+       print_error(_("Cannot find the current path %s in the directory tree.\n"), curPath);
+       return NULL;
+    }
     ydiff -= (wcd_cwin.curNode->y);
 
     if (wcd_cwin.graphics_mode & WCD_GRAPH_CENTER)
